@@ -10,7 +10,18 @@ module ActiveRecord
       # Generates all the attribute related methods for columns in the database
       # accessors, mutators and query methods.
       def define_attribute_methods
-        super(columns_hash.keys)
+        return if attribute_methods_generated?
+        super(column_names)
+        @attribute_methods_generated = true
+      end
+
+      def attribute_methods_generated?
+        @attribute_methods_generated ||= false
+      end
+
+      def undefine_attribute_methods(*args)
+        super
+        @attribute_methods_generated = false
       end
 
       # Checks whether the method is defined in the model or any of its subclasses
@@ -18,7 +29,11 @@ module ActiveRecord
       # method is defined by Active Record though.
       def instance_method_already_implemented?(method_name)
         method_name = method_name.to_s
-        @_defined_class_methods         ||= ancestors.first(ancestors.index(ActiveRecord::Base)).sum([]) { |m| m.public_instance_methods(false) | m.private_instance_methods(false) | m.protected_instance_methods(false) }.map {|m| m.to_s }.to_set
+        index = ancestors.index(ActiveRecord::Base) || ancestors.length
+        @_defined_class_methods         ||= ancestors.first(index).map { |m|
+          m.instance_methods(false) | m.private_instance_methods(false)
+        }.flatten.map {|m| m.to_s }.to_set
+
         @@_defined_activerecord_methods ||= defined_activerecord_methods
         raise DangerousAttributeError, "#{method_name} is defined by ActiveRecord" if @@_defined_activerecord_methods.include?(method_name)
         @_defined_class_methods.include?(method_name)
@@ -27,9 +42,8 @@ module ActiveRecord
       def defined_activerecord_methods
         active_record = ActiveRecord::Base
         super_klass   = ActiveRecord::Base.superclass
-        methods =  active_record.public_instance_methods - super_klass.public_instance_methods
-        methods += active_record.private_instance_methods - super_klass.private_instance_methods
-        methods += active_record.protected_instance_methods - super_klass.protected_instance_methods
+        methods = (active_record.instance_methods - super_klass.instance_methods) +
+                  (active_record.private_instance_methods - super_klass.private_instance_methods)
         methods.map {|m| m.to_s }.to_set
       end
     end
@@ -47,14 +61,14 @@ module ActiveRecord
       end
     end
 
-    def respond_to?(*args)
-      self.class.define_attribute_methods
+    def respond_to?(name, include_private = false)
+      self.class.define_attribute_methods unless self.class.attribute_methods_generated?
       super
     end
 
     protected
       def attribute_method?(attr_name)
-        attr_name == 'id' || attributes.include?(attr_name)
+        attr_name == 'id' || (defined?(@attributes) && @attributes.include?(attr_name))
       end
   end
 end
