@@ -13,18 +13,19 @@ module ActiveRecord
 
       # Returns a hash of attributes before typecasting and deserialization.
       def attributes_before_type_cast
-        Hash[attribute_names.map { |name| [name, read_attribute_before_type_cast(name)] }]
+        @attributes
       end
 
       private
-        # Handle *_before_type_cast for method_missing.
-        def attribute_before_type_cast(attribute_name)
-          if attribute_name == 'id'
-            read_attribute_before_type_cast(self.class.primary_key)
-          else
-            read_attribute_before_type_cast(attribute_name)
-          end
+
+      # Handle *_before_type_cast for method_missing.
+      def attribute_before_type_cast(attribute_name)
+        if attribute_name == 'id'
+          read_attribute_before_type_cast(self.class.primary_key)
+        else
+          read_attribute_before_type_cast(attribute_name)
         end
+      end
     end
   end
 end
