@@ -10,7 +10,13 @@ module ActiveRecord
       module ClassMethods
         protected
           def define_method_attribute=(attr_name)
-            generated_attribute_methods.module_eval("def #{attr_name}=(new_value); write_attribute('#{attr_name}', new_value); end", __FILE__, __LINE__)
+            if attr_name =~ ActiveModel::AttributeMethods::COMPILABLE_REGEXP
+              generated_attribute_methods.module_eval("def #{attr_name}=(new_value); write_attribute('#{attr_name}', new_value); end", __FILE__, __LINE__)
+            else
+              generated_attribute_methods.send(:define_method, "#{attr_name}=") do |new_value|
+                write_attribute(attr_name, new_value)
+              end
+            end
           end
       end
 
@@ -26,6 +32,7 @@ module ActiveRecord
           @attributes[attr_name] = value
         end
       end
+      alias_method :raw_write_attribute, :write_attribute
 
       private
         # Handle *= for method_missing.
